import time
import threading
from musys import memchan

'''
    Time keeper class to handle sessions
'''
class chronos(threading.Thread):
    def __init__(self, begin, delta, chan):
        threading.Thread.__init__(self)
        self.begin = begin
        self.delta = delta
        self.signal = True
        self.chan = chan
        self.block = False
        self.clearing = False
        self.quit = False

    '''
        Reset timer
    '''
    def reset_timer(self):
        self.begin = time.time()

    '''
        Thread's primary loop
    '''
    def run(self):
        while self.signal:
            ctime = time.time()
            if ctime - self.begin > self.delta or self.quit:
                self.signal = False
                if not self.block:
                    self.clearing = True
                    self.chan.memory_clear()
                    self.clearing = False
            time.sleep(1)
            
'''
    Memory Session class. The primary passthrough class that manages session settings
'''
class memsess:
    def __init__(self, auth_token, port, local_id, session_length, reset_timeout=False, host=None):
        try:
            self.port = int(port)
            self.session_length = int(session_length)
            self.auth_token = str(auth_token)
            self.local_id = str(local_id)
        except ValueError as _err:
            logging.error(_err)
            raise
        if self.port < 1 or self.port > 65535:
            logging.error("Port number out of range")
            raise Exception("Invalid port")

        self.port = str(self.port)
        if host is not None:
            self.host = host
        else:
            self.host = "localhost"

        self.reset_timeout = reset_timeout
        
        self.chan = memchan(
            self.auth_token,
            self.port,
            self.local_id,
            host=self.host
            )

    '''
        Start a session
    '''
    def beginSession(self):
        self.begin = time.time()
        self.timer = chronos(
            self.begin, 
            self.session_length,
            self.chan
            )
        self.timer.daemon = True
        self.timer.start()

    '''
        Stop a session
    '''
    def closeSession(self):
        self.timer.quit = True
        self.timer.join()
        del self.timer
        self.begin = None

    '''
        Put data on remote machine
    '''
    def put(self, label, data):
        if self.begin is None:
            return None
        if self.reset_timeout:
            self.timer.reset_timer()
        if self.timer.clearing:
            return None
        self.timer.block = True
        return_data = self.chan.put(label, data)
        self.timer.block = False
        return return_data

    '''
        Fetch data from a remote machine
    '''
    def fetch(self, label):
        if self.begin is None:
            return None
        if self.reset_timeout:
            self.timer.reset_timer()
        if self.timer.clearing:
            return None
        self.timer.block = True
        return_data =  self.chan.fetch(label)
        self.timer.block = False
        return return_data


import logging
import threading
from sys import getsizeof
from xmlrpc.server import SimpleXMLRPCServer

'''
    Network Memory
    A network-attatched memory module. For a module to be reached
    by external systems, ip must be set to the current device ip
    instead of the loopback address 
'''
class netmem(threading.Thread):
    def __init__(self, auth_token, port, host=None, max_storage=None):
        # Init thread
        threading.Thread.__init__(self)

        # Check variables and setup self
        try:
            if max_storage is not None:
                self.max_storage = int(max_storage)
            else:
                self.max_storage = None
            self.port = int(port)
            self.auth_token = str(auth_token)
        except ValueError as _err:
            logging.error(_err)
            raise
        if self.port < 1 or self.port > 65535:
            logging.error("Port number out of range")
            raise Exception("Invalid port")

        # Memory information
        # This will be a dict of dicts
        self.memory = {}
        self.lock = threading.Lock()

        self.errors = {
            "AUTH": "INVALID_AUTH_TOKEN",
            "CAP": "MEMORY_CAP_REACHED",
            "KEY": "MEMORY KEY NOT FOUND",
            "OKAY": "SUCCESS"
        }

        if host is None:
            self.host = "localhost"
        else:
            self.host = host

        # Create the XMLRPC server
        self.server = SimpleXMLRPCServer(
            (self.host, self.port), 
            logRequests=True
            )

        # Register Memory functions
        self.server.register_function(self.memory_put)
        self.server.register_function(self.memory_fetch)
        self.server.register_function(self.memory_clear)

    '''
        Start serving XMLRPC thread until self.signal 
        indicated close
    '''
    def run(self):
        self.signal = True
        while self.signal:
            self.server.handle_request()

    '''
        Generate return error
            - Non RPC
    '''
    def spawn_error(self, code):
        return "RETURN_STATUS:" + self.errors[code]

    '''
        Validate a request token
            - Non RPC
    '''
    def authorize_request(self, incoming_auth_token):
        if self.auth_token != incoming_auth_token:
            logging.warning("Unauthorized access attempt")
            return False
        return True

    '''
        Request to add something to memory
            - RPC
    '''
    def memory_put(self, auth_token, remote_id, label, data):
        if not self.authorize_request(auth_token):
            return (False, self.spawn_error("AUTH"))
    
        if self.max_storage is not None:
            new_mem_size = getsizeof(self.memory) + getsizeof(data)
            if new_mem_size >= self.max_storage:
                logging.warning("Memory cap reached")
                return (False, self.spawn_error("CAP"))

        self.lock.acquire()
        try:
            # Ensure that remote_id exists as a dict
            if remote_id not in self.memory:
                self.memory[remote_id] = {}

            # Set remote_id[label] to the data
            self.memory[remote_id][label] = data
        finally:
            self.lock.release()

        return (True, self.spawn_error("OKAY"))

    '''
        Request to get something from memory
            - RPC
    '''
    def memory_fetch(self, auth_token, remote_id, label):
        if not self.authorize_request(auth_token):
            return (False, self.spawn_error("AUTH"))

        self.lock.acquire()
        try:
            return_data = self.memory[remote_id][label]
        except KeyError:
            return_data = None
        finally:
            self.lock.release()

        if return_data is None:
            return (False, self.spawn_error("KEY"))

        return (True, return_data)
        
    '''
        Request to get remove a device's memory
            - RPC
    '''
    def memory_clear(self, auth_token, remote_id):
        if not self.authorize_request(auth_token):
            return (False, self.spawn_error("AUTH"))

        self.lock.acquire()
        try:
            if remote_id in self.memory:
                logging.info("Removing a device's data")
                del self.memory[remote_id]
                return_data = self.spawn_error("OKAY")
            else:
                return_data = self.spawn_error("KEY")
        finally:
            self.lock.release()

        if return_data is None:
            return (False, self.spawn_error("KEY"))

        return (True, return_data)



"""
musys
--------

musys is an XML-RPC memory storage application

```````````````

::

import musys



And Easy to Install
```````````````````

::

    $ pip3 install musys


"""


from setuptools import setup, find_packages
setup(
    name="muSys",
    version="0.1.1",
    description="A simple network attatched memory module",
    author="Josh Bosley",
    author_email="bosley117@gmail.com",
    license="MIT",
    classifiers = [
        "Programming Language :: Python",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Developers",
        "Topic :: Database" ],
    packages=find_packages(),
    install_requires=['dill']
    )



import dill
import logging
import xmlrpc.client

'''
    Memory Channel - For interacting with network-based memory
'''
class memchan:
    def __init__(self, auth_token, port, local_id, host=None):
        try:
            self.port = int(port)
            self.auth_token = str(auth_token)
            self.local_id = str(local_id)
        except ValueError as _err:
            logging.error(_err)
            raise
        if self.port < 1 or self.port > 65535:
            logging.error("Port number out of range")
            raise Exception("Invalid port")

        self.port = str(self.port)
        if host is not None:
            self.host = "http://"+host+":"+self.port
        else:
            self.host = "http://localhost:"+self.port
        
        self.accessor = xmlrpc.client.ServerProxy(self.host)

    '''
        Store data on the network attatched memory
    '''
    def put(self, label, data):
        if type(label) != str:
            logging.error("Label must be type 'string'")
            return None
        
        if type(data) != bytes:
            data = dill.dumps(data)

        result = self.accessor.memory_put(
            self.auth_token, 
            self.local_id,
            label,
            data
            )
        
        if not result[0]:
            logging.error(result[1])
            return None
        else:
            return True
        
    '''
        Get data from the network attatched memory
    '''
    def fetch(self, label):
        if type(label) != str:
            logging.error("Label must be type 'string'")
            return None

        result = self.accessor.memory_fetch(
            self.auth_token,
            self.local_id,
            label
        )

        if not result[0]:
            logging.error(result[1])
            return None
        else:
            # XMLRPC takes the dilled data and bins it,
            # so we have to .data it before loads
            return dill.loads(result[1].data) 

    '''
        Delete memory belonging to this instance 
        from the network attatched memory
    '''
    def memory_clear(self):
        self.accessor.memory_clear(
            self.auth_token,
            self.local_id
        )
